import asyncio
import joblib
from typing import List, Dict, Any
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.preprocessing import MultiLabelBinarizer
from sklearn.multiclass import OneVsRestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.pipeline import Pipeline


class ToolCallPredictor:
    def __init__(self, model_path: str = "toolcall_model.pkl"):
        # Путь к файлу, где будет сохранена обученная модель и бинаризатор
        self.model_path = model_path
        self.pipeline: Pipeline | None = None  # ML-пайплайн (Tfidf + OneVsRestClassifier)
        self.label_binarizer: MultiLabelBinarizer | None = None  # Бинаризатор мульти-меток

    async def init_from_collector(self, collector) -> None:
        """Автоматически обучает модель по descriptions из collector."""
        tools = await collector.get_tools()
        data = []
        for tool in tools:
            description = tool["function"].get("description", "")
            name = tool["function"].get("name", "")
            if description and name:
                data.append({
                    "text": description,
                    "tools": [name]
                })

        if not data:
            print("[ToolCallPredictor] ⚠️ Нет descriptions в collector — пропускаем init_from_collector")
            return

        print(f"[ToolCallPredictor] ⚙️ Инициализация модели по {len(data)} tools из collector...")
        await self.train(data)
        await self.load_model()
        print("[ToolCallPredictor] ✅ Init from collector завершён")

    async def train(self, data: List[Dict[str, Any]]) -> None:
        """Асинхронно обучает модель на размеченных данных."""
        texts = [item["text"] for item in data]  # Список текстов
        labels = [item["tools"] for item in data]  # Список списков инструментов

        # Преобразуем список списков в бинарную матрицу
        self.label_binarizer = MultiLabelBinarizer()
        y = self.label_binarizer.fit_transform(labels)

        # Создаем pipeline: TF-IDF векторизация + логистическая регрессия в One-vs-Rest схеме
        pipeline = Pipeline([
            ('vectorizer', TfidfVectorizer()),
            ('classifier', OneVsRestClassifier(LogisticRegression(solver="liblinear")))
        ])

        # Обучаем модель в отдельном потоке, чтобы не блокировать event loop
        self.pipeline = await asyncio.to_thread(pipeline.fit, texts, y)

        # Сохраняем модель и бинаризатор в файл
        await asyncio.to_thread(joblib.dump, (self.pipeline, self.label_binarizer), self.model_path)

    async def load_model(self) -> None:
        """Загружает обученную модель и бинаризатор из файла."""
        self.pipeline, self.label_binarizer = await asyncio.to_thread(joblib.load, self.model_path)

    async def predict_tools(self, text: str) -> List[str]:
        """Предсказывает список названий инструментов для одного текста."""
        if not self.pipeline or not self.label_binarizer:
            raise RuntimeError("Модель не загружена")

        prediction = await asyncio.to_thread(self.pipeline.predict, [text])
        tools = self.label_binarizer.inverse_transform(prediction)[0]
        return list(tools)

    async def predict_tools_batch(self, texts: List[str]) -> List[List[str]]:
        """Пакетное предсказание списков инструментов для каждого текста."""
        if not self.pipeline or not self.label_binarizer:
            raise RuntimeError("Модель не загружена")

        predictions = await asyncio.to_thread(self.pipeline.predict, texts)
        return self.label_binarizer.inverse_transform(predictions)

    async def generate_tool_calls(self, text: str) -> List[Dict[str, Any]]:
        """Генерирует tool_calls (название + аргументы) для одного текста."""
        tools = await self.predict_tools(text)
        return [
            {
                "name": tool_name,
                "arguments": {"query": text}  # Можно заменить на кастомную логику
            } for tool_name in tools
        ]

    async def generate_tool_calls_batch(self, texts: List[str]) -> List[List[Dict[str, Any]]]:
        """Пакетная генерация tool_calls для каждого текста."""
        all_tools = await self.predict_tools_batch(texts)
        return [
            [
                {
                    "name": tool_name,
                    "arguments": {"query": text}  # Можно заменить на продвинутый парсинг
                }
                for tool_name in tools
            ]
            for text, tools in zip(texts, all_tools)
        ]

    async def prepare_user_message(self, text: str) -> Dict[str, Any]:
        """Формирует одно сообщение пользователя для Assistant API с tool_calls."""
        tool_calls = await self.generate_tool_calls(text)
        return {
            "role": "user",
            "content": text,
            "tool_calls": tool_calls
        }

    async def prepare_user_messages_batch(self, texts: List[str]) -> List[Dict[str, Any]]:
        """Формирует список сообщений для Assistant API для каждого входного текста."""
        tool_calls_list = await self.generate_tool_calls_batch(texts)
        return [
            {
                "role": "user",
                "content": text,
                "tool_calls": tool_calls
            }
            for text, tool_calls in zip(texts, tool_calls_list)
        ]



import importlib.util
from typing import List, Dict, Any
from .tool_call_predictor import ToolCallPredictor
from .tool_dispatcher import ToolDispatcher
from .tools_collector import ToolsCollector
from .file_manager import StorageManager
from pathlib import Path


class ToolManager:
    def __init__(
        self,
        assistant_id: str,
        user_id: str,
        predictor: ToolCallPredictor,
        storage: StorageManager,
        collector: ToolsCollector,
        tool_module_path: str = "custom_tools.py"
    ):
        self.assistant_id = assistant_id
        self.user_id = user_id
        self.predictor = predictor
        self.collector = collector
        self.storage = storage
        self.dispatcher = ToolDispatcher(
            assistant_id=assistant_id,
            user_id=user_id,
            storage_manager=storage,
            tools_collector=collector
        )
        self._load_custom_tools(tool_module_path)

    def _load_custom_tools(self, module_path: str):
        """
        Импортирует модуль с инструментами, регистрируя все tool-функции.
        """
        module_path = Path(module_path)
        if not module_path.exists():
            print(f"[ToolManager] ⚠️ Файл {module_path} не найден — инструменты не загружены.")
            return

        spec = importlib.util.spec_from_file_location("custom_tools", str(module_path))
        if spec and spec.loader:
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            print(f"[ToolManager] ✅ Инструменты загружены из {module_path}")
        else:
            print(f"[ToolManager] ❌ Не удалось загрузить модуль инструментов из {module_path}")

    async def generate_tool_calls(self, text: str) -> List[Dict[str, Any]]:
        return await self.predictor.generate_tool_calls(text)

    async def generate_user_message(self, text: str) -> Dict[str, Any]:
        return await self.predictor.prepare_user_message(text)

    async def handle_tool_calls(self, tool_calls: list) -> List[Dict[str, Any]]:
        return await self.dispatcher.dispatch(tool_calls)

    def get_generated_files(self) -> List[str]:
        return self.dispatcher.get_file_names()


from .file_manager import StorageManager
from .tools_collector import ToolsCollector, tools_collector
from .tool_call_predictor import ToolCallPredictor
from .tool_dispatcher import ToolDispatcher
from .tool_manager import ToolManager

__all__ = [
    "StorageManager",
    "ToolsCollector",
    "tools_collector",
    "ToolCallPredictor",
    "ToolDispatcher",
    "ToolManager",
]


from typing import Callable, Dict, List, Any
import inspect
import asyncio
import jsonschema



class ToolsCollector:
    def __init__(self):
        self._tools: Dict[str, Dict] = {}         # Словарь инструкций tool calling
        self._functions: Dict[str, Callable] = {} # Словарь самих функций

    def tool(self, schema: Dict) -> Callable:
        """
        Декоратор для регистрации tool-функции с инструкцией и схемой параметров.
        """
        def decorator(func: Callable) -> Callable:
            name = schema.get("function", {}).get("name", func.__name__)
            self._tools[name] = schema
            self._functions[name] = func
            return func
        return decorator

    async def get_tools(self) -> List[Dict]:
        """Возвращает все tool-схемы для Assistant API."""
        return list(self._tools.values())

    async def get_function(self, name: str) -> Callable:
        """Возвращает зарегистрированную функцию по имени."""
        if name not in self._functions:
            raise KeyError(f"Tool '{name}' не зарегистрирован.")
        return self._functions[name]

    async def get_all(self) -> Dict[str, Callable]:
        """Возвращает словарь всех зарегистрированных функций."""
        return self._functions

    async def call_tool(self, name: str, arguments: Dict[str, Any]) -> Any:
        """
        Асинхронно вызывает функцию по имени, валидируя входные аргументы по JSON Schema.
        Поддерживает как обычные, так и async-функции.
        """
        if name not in self._tools:
            raise KeyError(f"Tool '{name}' не зарегистрирован.")

        schema = self._tools[name]["function"]["parameters"]
        try:
            jsonschema.validate(instance=arguments, schema=schema)
        except jsonschema.exceptions.ValidationError as e:
            raise ValueError(f"Невалидные аргументы для '{name}': {e.message}")

        func = self._functions[name]
        if inspect.iscoroutinefunction(func):
            return await func(**arguments)
        else:
            return await asyncio.to_thread(func, **arguments)

# Глобальный экземпляр, импортируемый везде
tools_collector = ToolsCollector()


import json
from typing import List, Dict, Any
from .file_manager import StorageManager
from .tools_collector import ToolsCollector

OPENAI_SAFE_OUTPUT_CHUNK = 3000


class ToolDispatcher:
    def __init__(self, assistant_id: str, user_id: str, storage_manager: StorageManager, tools_collector: ToolsCollector):
        self.assistant_id = assistant_id
        self.user_id = user_id
        self.storage_manager = storage_manager
        self.tools_collector = tools_collector
        self.collected_file_names: List[str] = []

    def _split_output_text(self, text: str, max_len: int = OPENAI_SAFE_OUTPUT_CHUNK) -> List[str]:
        return [text[i:i + max_len] for i in range(0, len(text), max_len)]

    async def dispatch(self, tool_calls: List[Any]) -> List[Dict[str, Any]]:
        results = []

        for call in tool_calls:
            try:
                function_name = call.function.name
                arguments_str = call.function.arguments or "{}"

                try:
                    arguments = json.loads(arguments_str)
                except json.JSONDecodeError:
                    print(f"[ToolDispatcher] Ошибка парсинга JSON: {arguments_str}")
                    arguments = {}

                # Преобразование вложенных строковых JSON-полей в словари
                for key in [
                    "downloaded_all_files", "downloaded_media_files", "downloaded_text_files",
                    "downloaded_image", "downloaded_video", "downloaded_audio",
                    "downloaded_tables", "downloaded_documents", "downloaded_drawings",
                    "downloaded_archives", "actions_per_file"
                ]:
                    if key in arguments and isinstance(arguments[key], str):
                        try:
                            arguments[key] = json.loads(arguments[key].replace("'", '"'))
                        except Exception as e:
                            print(f"[ToolDispatcher] Не удалось преобразовать {key} в dict: {e}")
                            arguments[key] = {}

                try:
                    result = await self.tools_collector.call_tool(function_name, arguments)
                except Exception as tool_error:
                    results.append({
                        "tool_call_id": call.id,
                        "output": f"❌ Ошибка вызова инструмента '{function_name}': {tool_error}"
                    })
                    continue

                if isinstance(result, dict):
                    summary = str(result.get("summary", "✅ Операция выполнена успешно."))
                    message_content = result.get("message", {})

                    # Добавляем возвращённые пути к новым файлам
                    if "new_file_path" in result:
                        new_paths = result["new_file_path"]
                        if isinstance(new_paths, list):
                            self.collected_file_names.extend([str(p) for p in new_paths])

                    output_json_str = json.dumps(message_content, ensure_ascii=False, indent=2) \
                        if isinstance(message_content, dict) else str(message_content)

                    if len(output_json_str) > OPENAI_SAFE_OUTPUT_CHUNK:
                        chunks = self._split_output_text(output_json_str)
                        filename = f"result_{function_name}_{call.id}.json"

                        await self.storage_manager.save_file(
                            user_id=self.user_id,
                            assistant_id=self.assistant_id,
                            file_name=filename,
                            content="\n".join(chunks),
                            expiry_hours=48
                        )

                        self.collected_file_names.append(filename)

                        output_text = json.dumps({
                            "summary": summary,
                            "note": f"📎 Подробный результат сохранён как `{filename}` в папке ассистента.",
                            "filename": filename
                        }, ensure_ascii=False, indent=2)
                    else:
                        output_text = summary + "\n" + output_json_str

                    results.append({
                        "tool_call_id": call.id,
                        "output": output_text
                    })
                else:
                    results.append({
                        "tool_call_id": call.id,
                        "output": str(result)
                    })

            except Exception as e:
                results.append({
                    "tool_call_id": getattr(call, "id", "undefined"),
                    "output": f"❌ Общая ошибка при обработке: {e}"
                })

        return results

    def get_file_names(self) -> List[str]:
        return self.collected_file_names


import os
import json
import shutil
import asyncio
from datetime import datetime, timedelta
from pathlib import Path
from typing import List, Optional, Type
from pydantic import BaseModel, ValidationError


class StorageManager:
    def __init__(self, base_path: Path = Path("local_storage"), cleanup_interval: int = 3600):
        self.base_path = base_path
        self.cleanup_interval = cleanup_interval
        self._file_metadata: dict[Path, datetime] = {}
        self._lock = asyncio.Lock()
        self._cleanup_task = None  # Таск теперь НЕ запускаем в __init__

    async def start_cleanup(self):
        """Запустить фоновую задачу очистки."""
        if self._cleanup_task is None:
            self._cleanup_task = asyncio.create_task(self._auto_cleanup())

    async def _auto_cleanup(self):
        while True:
            await asyncio.sleep(self.cleanup_interval)
            await self._cleanup()

    async def _cleanup(self):
        now = datetime.utcnow()
        async with self._lock:
            to_delete = []
            for path, expiry in list(self._file_metadata.items()):
                if now > expiry and path.exists():
                    try:
                        path.unlink()
                        to_delete.append(path)
                    except Exception as e:
                        print(f"Ошибка при удалении {path}: {e}")
            for path in to_delete:
                del self._file_metadata[path]

    async def save_file(
        self,
        user_id: str,
        assistant_id: str,
        file_name: str,
        content: str,
        expiry_hours: int,
        schema: Optional[Type[BaseModel]] = None
    ) -> Path:
        file_path = self.base_path / user_id / assistant_id / file_name
        file_path.parent.mkdir(parents=True, exist_ok=True)

        # Валидация JSON-файлов по Pydantic-схеме
        if file_path.suffix == ".json" and schema is not None:
            try:
                parsed = json.loads(content)
                schema.parse_obj(parsed)
            except (json.JSONDecodeError, ValidationError) as e:
                raise ValueError(f"Файл не прошёл валидацию: {e}")

        file_path.write_text(content, encoding="utf-8")
        expiry_time = datetime.utcnow() + timedelta(hours=expiry_hours)

        async with self._lock:
            self._file_metadata[file_path] = expiry_time

        return file_path

    async def delete_files(self, user_id: str, assistant_id: str, file_names: List[str]) -> None:
        for name in file_names:
            file_path = self.base_path / user_id / assistant_id / name
            if file_path.exists():
                try:
                    file_path.unlink()
                    async with self._lock:
                        self._file_metadata.pop(file_path, None)
                except Exception as e:
                    print(f"Ошибка при удалении {file_path}: {e}")

    async def list_files(self, user_id: str, assistant_id: str) -> List[str]:
        dir_path = self.base_path / user_id / assistant_id
        if not dir_path.exists():
            return []
        return [f.name for f in dir_path.iterdir() if f.is_file()]

    async def delete_assistant_folder(self, user_id: str, assistant_id: str) -> None:
        folder_path = self.base_path / user_id / assistant_id
        if folder_path.exists():
            shutil.rmtree(folder_path)
            async with self._lock:
                self._file_metadata = {
                    path: exp for path, exp in self._file_metadata.items()
                    if not str(path).startswith(str(folder_path))
                }

    async def get_file_path(self, user_id: str, assistant_id: str, file_name: str) -> Optional[Path]:
        file_path = self.base_path / user_id / assistant_id / file_name
        if file_path.exists():
            return file_path
        return None

    async def get_files_by_extension(self, user_id: str, assistant_id: str, extension: str) -> List[Path]:
        dir_path = self.base_path / user_id / assistant_id
        if not dir_path.exists():
            return []
        return [f for f in dir_path.iterdir() if f.is_file() and f.suffix == extension]




import pytest
import asyncio
import os
from ai_agent.tools.file_manager import StorageManager
from pydantic import BaseModel

class DummySchema(BaseModel):
    name: str
    value: int

@pytest.mark.asyncio
async def test_save_and_list_files(tmp_path):
    sm = StorageManager(base_path=tmp_path)
    path = await sm.save_file("user1", "agent1", "test.txt", "hello", 1)
    files = await sm.list_files("user1", "agent1")
    assert "test.txt" in files

@pytest.mark.asyncio
async def test_save_json_with_schema(tmp_path):
    sm = StorageManager(base_path=tmp_path)
    valid_json = '{"name": "test", "value": 123}'
    path = await sm.save_file("user1", "agent1", "data.json", valid_json, 1, schema=DummySchema)
    assert path.exists()

@pytest.mark.asyncio
async def test_delete_files(tmp_path):
    sm = StorageManager(base_path=tmp_path)
    await sm.save_file("user1", "agent1", "delete.txt", "data", 1)
    await sm.delete_files("user1", "agent1", ["delete.txt"])
    files = await sm.list_files("user1", "agent1")
    assert "delete.txt" not in files

@pytest.mark.asyncio
async def test_get_file_path(tmp_path):
    sm = StorageManager(base_path=tmp_path)
    await sm.save_file("user1", "agent1", "find.txt", "content", 1)
    path = await sm.get_file_path("user1", "agent1", "find.txt")
    assert path is not None

@pytest.mark.asyncio
async def test_get_files_by_extension(tmp_path):
    sm = StorageManager(base_path=tmp_path)
    await sm.save_file("user1", "agent1", "a.txt", "1", 1)
    await sm.save_file("user1", "agent1", "b.log", "2", 1)
    txt_files = await sm.get_files_by_extension("user1", "agent1", ".txt")
    assert len(txt_files) == 1



import asyncio
import joblib
from typing import List, Dict, Any
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.preprocessing import MultiLabelBinarizer
from sklearn.multiclass import OneVsRestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.pipeline import Pipeline


class ToolCallPredictor:
    def __init__(self, model_path: str = "toolcall_model.pkl"):
        self.model_path = model_path
        self.pipeline: Pipeline | None = None
        self.label_binarizer: MultiLabelBinarizer | None = None

    async def train(self, data: List[Dict[str, Any]]) -> None:
        """Асинхронно обучает модель на размеченных данных"""
        texts = [item["text"] for item in data]
        labels = [item["tools"] for item in data]

        self.label_binarizer = MultiLabelBinarizer()
        y = self.label_binarizer.fit_transform(labels)

        pipeline = Pipeline([
            ('vectorizer', TfidfVectorizer()),
            ('classifier', OneVsRestClassifier(LogisticRegression(solver="liblinear")))
        ])

        # Обучение модели в отдельном потоке
        self.pipeline = await asyncio.to_thread(pipeline.fit, texts, y)

        # Сохранение модели и бинализатора
        await asyncio.to_thread(joblib.dump, (self.pipeline, self.label_binarizer), self.model_path)

    async def load_model(self) -> None:
        """Загружает модель и бинализатор из файла"""
        self.pipeline, self.label_binarizer = await asyncio.to_thread(joblib.load, self.model_path)

    async def predict_tools(self, text: str) -> List[str]:
        """Предсказывает список tool names на основе текста"""
        if not self.pipeline or not self.label_binarizer:
            raise RuntimeError("Модель не загружена")

        prediction = await asyncio.to_thread(self.pipeline.predict, [text])
        tools = self.label_binarizer.inverse_transform(prediction)[0]
        return list(tools)

    async def generate_tool_calls(self, text: str) -> List[Dict[str, Any]]:
        """Формирует список toolcalls в нужном формате"""
        tools = await self.predict_tools(text)
        return [
            {
                "name": tool_name,
                "arguments": {"query": text}  # Здесь может быть более умная логика
            } for tool_name in tools
        ]

    async def prepare_user_message(self, text: str) -> Dict[str, Any]:
        """Генерирует финальное сообщение для Assistant API"""
        tool_calls = await self.generate_tool_calls(text)
        return {
            "role": "user",
            "content": text,
            "tool_calls": tool_calls
        }



import pytest
from unittest.mock import AsyncMock, MagicMock
from ai_agent.tools.tool_dispatcher import ToolDispatcher
from ai_agent.tools.file_manager import StorageManager
from ai_agent.tools.tools_collector import ToolsCollector

@pytest.mark.asyncio
async def test_dispatch_no_calls(tmp_path):
    dispatcher = ToolDispatcher(
        assistant_id="aid", user_id="uid",
        storage_manager=StorageManager(base_path=tmp_path),
        tools_collector=ToolsCollector()
    )
    result = await dispatcher.dispatch([])
    assert result == []



import pytest
from unittest.mock import AsyncMock

@pytest.mark.asyncio
async def test_create(gpt_agent):
    assistant_id, thread_id = await gpt_agent.create()
    assert assistant_id == "assistant-123"
    assert thread_id == "thread-456"
    assert gpt_agent.initialized is True

@pytest.mark.asyncio
async def test_update_agent(gpt_agent):
    await gpt_agent.create()
    response = await gpt_agent.update_agent(
        new_assistant_name="NewName",
        new_assistant_instructions="New instructions",
        new_assistant_model="gpt-4o"
    )
    assert gpt_agent.name == "NewName"
    assert gpt_agent.instructions == "New instructions"
    assert gpt_agent.model == "gpt-4o"

@pytest.mark.asyncio
async def test_update_memory(gpt_agent):
    await gpt_agent.create()
    new_thread_id = await gpt_agent.update_memory()
    assert new_thread_id == "thread-456"

@pytest.mark.asyncio
async def test_send_text_message(gpt_agent):
    await gpt_agent.create()
    run_id = await gpt_agent.send_text_message(text="Hello, world!")
    assert run_id == "run-789"

@pytest.mark.asyncio
async def test_send_image_message(gpt_agent):
    await gpt_agent.create()
    gpt_agent.get_image_description = AsyncMock(return_value="Test image description")
    run_id = await gpt_agent.send_image_message(text="Hello image", file_path=["test.jpg"])
    assert run_id == "run-789"

@pytest.mark.asyncio
async def test_message_text(gpt_agent):
    await gpt_agent.create()
    await gpt_agent.message(text="Test text", file_path=None)

@pytest.mark.asyncio
async def test_message_image(gpt_agent):
    await gpt_agent.create()
    gpt_agent.get_image_description = AsyncMock(return_value="Test image description")
    await gpt_agent.message(text="Test image", file_path=["test.jpg"])

@pytest.mark.asyncio
async def test_get_image_description(gpt_agent):
    await gpt_agent.create()
    gpt_agent.get_image_description = AsyncMock(return_value="Test image description")
    description = await gpt_agent.get_image_description("fake_image.jpg")
    assert description == "Test image description"


import os
import base64
import openai
import asyncio
from ai_agent.tools import ToolManager, ToolCallPredictor, StorageManager, tools_collector

# Функция для кодирования изображения в Base64
def encode_image(image_path):
    with open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode("utf-8")

class GptAgent:
    def __init__(self, openai_key, instructions: str, name: str, model: str, assistant_id=None, thread_id: str = None):
        self.client = openai.AsyncOpenAI(api_key=openai_key)
        self.instructions = instructions
        self.name = name
        self.model = model
        self.assistant_id = assistant_id
        self.thread_id = thread_id
        self.initialized = False  # Флаг инициализации

        # ✅ Инициализация ToolManager
        self.tool_manager = ToolManager(
            assistant_id=self.assistant_id or "pending",  # пока не создано — заглушка
            user_id="user-001",
            predictor=ToolCallPredictor("toolcall_model.pkl"),
            storage=StorageManager(),
            collector=tools_collector
        )

    async def create(self):
        assistant = await self.client.beta.assistants.create(
            name=self.name,
            instructions=self.instructions,
            model=self.model
        )
        self.assistant_id = assistant.id
        thread = await self.client.beta.threads.create()
        self.thread_id = thread.id
        self.initialized = True

        await self.tool_manager.predictor.init_from_collector(self.tool_manager.collector)

        # Обновляем ToolManager на правильный assistant_id
        self.tool_manager.assistant_id = self.assistant_id
        self.tool_manager.dispatcher.assistant_id = self.assistant_id

        return assistant.id, thread.id

    async def update_agent(self, new_assistant_name: str = None, new_assistant_instructions: str = None, new_assistant_model: str = None):
        l_name = self.name
        l_instructions = self.instructions
        l_model = self.model

        if new_assistant_name and new_assistant_name != self.name:
            l_name = new_assistant_name
        if new_assistant_instructions and new_assistant_instructions != self.instructions:
            l_instructions = new_assistant_instructions
        if new_assistant_model and new_assistant_model != self.model:
            l_model = new_assistant_model

        response = await self.client.beta.assistants.update(
            assistant_id=self.assistant_id,
            name=l_name,
            instructions=l_instructions,
            model=l_model
        )
        self.name = l_name
        self.instructions = l_instructions
        self.model = l_model

        return response

    async def update_memory(self):
        thread = await self.client.beta.threads.create()
        if thread.id and thread.id != self.thread_id:
            await self.client.beta.threads.delete(thread_id=self.thread_id)
            self.thread_id = thread.id
        return thread.id

    async def message(self, text: str, file_path: list[str] = None):
        # Если агент ещё не создан — создаём
        if not self.initialized and not (self.assistant_id and self.thread_id):
            await self.create()

        # Дождаться завершения последнего run (если есть)
        last_runs = await self.client.beta.threads.runs.list(thread_id=self.thread_id, limit=1)
        if last_runs.data:
            while True:
                run_status = await self.client.beta.threads.runs.retrieve(
                    thread_id=self.thread_id,
                    run_id=last_runs.data[0].id
                )
                if run_status.status == "completed":
                    break
                else:
                    await asyncio.sleep(1)

        # Определение расширений файлов
        image_extensions = {".png", ".jpg", ".jpeg", ".gif", ".bmp", ".webp"}
        is_image = False

        if file_path:
            for path in file_path:
                ext = os.path.splitext(path)[1].lower()
                if ext in image_extensions:
                    is_image = True
                    break

        # Отправляем сообщение в зависимости от типа
        if is_image:
            run_id = await self.send_image_message(text=text, file_path=file_path)
        else:
            run_id = await self.send_text_message(text=text, file_path=file_path)

        # Цикл обработки run (встроенная обработка tool_calls через ToolManager)
        while True:
            run = await self.client.beta.threads.runs.retrieve(thread_id=self.thread_id, run_id=run_id)
            if run.status == "completed":
                print("[GptAgent] ✅ Run completed.")
                break
            elif run.status == "requires_action":
                if run.required_action.type == "submit_tool_outputs":
                    tool_calls = run.required_action.submit_tool_outputs.tool_calls
                    print(f"[GptAgent] ⚙️ Handling {len(tool_calls)} tool_calls...")

                    # Используем ToolManager для обработки tool_calls
                    tool_outputs = await self.tool_manager.handle_tool_calls(tool_calls)

                    # Отправляем результаты обратно в run
                    await self.client.beta.threads.runs.submit_tool_outputs(
                        thread_id=self.thread_id,
                        run_id=run_id,
                        tool_outputs=tool_outputs
                    )
                    print("[GptAgent] ✅ Tool outputs submitted.")
            else:
                await asyncio.sleep(1)

    async def send_text_message(self, text: str, file_path: list[str] = None):
        file_descriptions = []
        if file_path:
            for path in file_path:
                filename = os.path.basename(path)
                file_descriptions.append(f"\nfile_name: '{filename}'")

        if file_descriptions:
            files_str = ", ".join(file_descriptions)
            prompt = text + f"\nИмена документов (файлов): {files_str}"
        else:
            prompt = text

        await self.client.beta.threads.messages.create(thread_id=self.thread_id, role="user", content=prompt)
        run = await self.client.beta.threads.runs.create(
            thread_id=self.thread_id,
            assistant_id=self.assistant_id
        )
        return run.id

    async def send_image_message(self, text: str, file_path: list[str] = None):
        image_list = []
        for idx, path in enumerate(file_path):
            filename = os.path.basename(path)
            path_description = await self.get_image_description(image_path=path)
            image_string = f"{idx+1}) image_name: '{filename}', image_description: '{path_description}'"
            image_list.append(image_string)

        complex_image_string = "\n".join(image_list)
        prompt = text + "\n" + complex_image_string

        await self.client.beta.threads.messages.create(thread_id=self.thread_id, role="user", content=prompt)
        run = await self.client.beta.threads.runs.create(
            thread_id=self.thread_id,
            assistant_id=self.assistant_id
        )
        return run.id

    async def get_image_description(self, image_path: str):
        # Кодируем изображение в Base64
        base64_image = encode_image(image_path)

        # Выполняем запрос к OpenAI API для обработки изображения
        response = await self.client.chat.completions.create(
            model="gpt-4o",
            messages=[{
                "role": "user",
                "content": [
                    {"type": "text", "text": "Что на изображении?"},
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": f"data:image/jpeg;base64,{base64_image}",
                        },
                    },
                ],
            }]
        )

        # Получаем описание изображения от модели
        image_description = response.choices[0].message.content.strip()
        return image_description




import pytest
import asyncio
import os
from ai_agent.tools.tool_call_predictor import ToolCallPredictor

@pytest.mark.asyncio
async def test_train_and_predict(tmp_path):
    model_path = tmp_path / "model.pkl"
    predictor = ToolCallPredictor(model_path=str(model_path))
    data = [
        {"text": "analyze image", "tools": ["analyzer"]},
        {"text": "translate text", "tools": ["translator"]},
    ]
    await predictor.train(data)
    await predictor.load_model()
    tools = await predictor.predict_tools("translate text")
    assert "translator" in tools



import pytest
from unittest.mock import AsyncMock
from ai_agent.tools.tool_manager import ToolManager
from ai_agent.tools.tool_call_predictor import ToolCallPredictor
from ai_agent.tools.file_manager import StorageManager
from ai_agent.tools.tools_collector import ToolsCollector

@pytest.mark.asyncio
async def test_generate_user_message(tmp_path):
    predictor = ToolCallPredictor()
    predictor.pipeline = None
    predictor.label_binarizer = None
    predictor.generate_tool_calls = AsyncMock(return_value=[])
    predictor.prepare_user_message = AsyncMock(return_value={"role": "user", "content": "hello", "tool_calls": []})

    manager = ToolManager(
        assistant_id="aid",
        user_id="uid",
        predictor=predictor,
        storage=StorageManager(base_path=tmp_path),
        collector=ToolsCollector()
    )
    msg = await manager.generate_user_message("hello")
    assert msg["role"] == "user"




import pytest
import asyncio
from ai_agent.tools.tools_collector import ToolsCollector

@pytest.mark.asyncio
async def test_tool_registration_and_call():
    collector = ToolsCollector()

    @collector.tool(schema={
        "function": {
            "name": "test_tool",
            "parameters": {
                "type": "object",
                "properties": {
                    "x": {"type": "integer"}
                },
                "required": ["x"]
            }
        }
    })
    def test_tool(x):
        return {"result": x * 2}

    tools = await collector.get_tools()
    assert any(t["function"]["name"] == "test_tool" for t in tools)

    func = await collector.get_function("test_tool")
    assert callable(func)

    result = await collector.call_tool("test_tool", {"x": 5})
    assert result["result"] == 10
